import json
import argparse
import logging

import rita


logger = logging.getLogger(__name__)


def main():
    parser = argparse.ArgumentParser(
        description="Compile rita -> spaCy patterns"
    )

    parser.add_argument("-f", help=".rita rules file")
    parser.add_argument(
        "out",
        help="output .jsonl file to store rules"
    )
    parser.add_argument("--debug", help="debug mode", action="store_true")
    parser.add_argument("--engine", help="Engine to use when compiling rules", default="spacy")
    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    patterns = rita.compile(args.f, use_engine=args.engine)

    logger.info("Compiling rules using {} engine".format(args.engine))

    with open(args.out, "w") as f:
        for pattern in patterns:
            f.write(json.dumps(pattern) + "\n")



import os

import logging
import types

from rita.config import with_config
from rita.preprocess import preprocess_rules


logger = logging.getLogger(__name__)

__version__ = (0, 4, 4, os.getenv("VERSION_PATCH"))


def get_version():
    normalized = list([i for i in __version__ if i is not None])
    if len(normalized) == 4:
        return "{0}.{1}.{2}-{3}".format(*normalized)
    else:
        return "{0}.{1}.{2}".format(*normalized)


@with_config
def compile_string(raw, config, use_engine=None, **kwargs):
    from rita.parser import RitaParser
    for k, v in kwargs.items():
        config.set_variable(k, v)

    parser = RitaParser(config)
    parser.build()
    root = parser.parse(raw)
    logger.debug(root)
    if use_engine:
        compile_rules = config.get_engine(use_engine)
    else:
        compile_rules = config.default_engine
    rules = list(preprocess_rules(root, config))
    result = compile_rules(rules, config)
    if isinstance(result, types.GeneratorType):
        return list(result)
    else:
        return result


def compile(fname, use_engine=None, **kwargs):
    with open(fname, "r") as f:
        raw = f.read()

    return compile_string(raw, use_engine=use_engine, **kwargs)



import logging

from unicodedata import normalize, category
from itertools import cycle

logger = logging.getLogger(__name__)


class Node(object):
    """
    An utility structure. Has no meaning outside
    Allows to specify single path showing how it branches
    and by doing `unwrap` we get multiple lists for each possible variation
    """
    def __init__(self, data=None):
        self.data = data
        self.children = []
        self.next_node = None
        self.children_cycle = None
        self.ref_count = 0
        self.depth = 0

    def add_child(self, c):
        self.children.append(Node(c))
        self.reset_cycle()

    def add_next(self, node):
        self.next_node = node

    @property
    def child(self):
        # Corner case of 0 depth
        if self.depth == 0:
            result = self.current
            self.next_child()
            return result

        if self.ref_count >= self.depth:
            self.next_child()
            self.ref_count = 0
        else:
            self.ref_count += 1
        return self.current

    def next_child(self):
        self.current = next(self.children_cycle)

    def reset_cycle(self):
        self.children_cycle = cycle(self.children)
        self.current = next(self.children_cycle)

    def unwrap(self):
        variants = 1
        current = self
        while current is not None:
            variants *= current.weight
            current = current.next_node

        logger.debug("Total variants: {}".format(variants))

        for i in range(0, variants):
            result = []
            current = self
            while current is not None:
                if current.data:
                    result.append(current.data)
                if len(current.children) > 0:
                    c = current.child
                    result.append(c.data)
                current = current.next_node
            yield result

    @property
    def weight(self):
        if len(self.children) == 0:
            return 1
        else:
            return len(self.children)

    def __repr__(self):
        return "{data}[{children}] -> {next_node}".format(
            data=self.data,
            children=", ".join(map(str, self.children)),
            next_node=str(self.next_node)
        )


class SingletonMixin(object):
    _instance = None

    def __new__(class_, *args, **kwargs):
        if not isinstance(class_._instance, class_):
            class_._instance = object.__new__(class_, *args, **kwargs)
        return class_._instance


def deaccent(text):
    return normalize("NFC",
                     "".join(c
                             for c in normalize("NFD", text)
                             if category(c) != "Mn"))



import logging

from functools import reduce

from rita.utils import Node, deaccent

logger = logging.getLogger(__name__)


def handle_prefix(rules, config):
    """
    If we have a prefix and rule following it, we apply this prefix on that rule
    """
    def apply_prefix(pattern, prefix):
        (name, args, op) = pattern
        if name == "any_of":
            return (name, list(["{0}{1}".format(prefix, item)
                                for item in args]), op)
        elif name == "value":
            return (name, "{0}{1}".format(prefix, args), op)
        else:
            logger.warning("Don't know how to apply prefix on: {}".format(name))
            return pattern

    def gen():
        prefix = None
        for p in pattern:
            (name, args, op) = p
            if name == "prefix":
                prefix = args
            else:
                if prefix:
                    yield apply_prefix(p, prefix)
                    prefix = None
                else:
                    yield p
    for group_label, pattern in rules:
        yield (group_label, list(gen()))


def handle_deaccent(rules, config):
    """
    If we get accented word, eg: {WORD("naïve"), WORD("bayes")}
    In case of word, it should become list => {IN_LIST({"naïve", "naive"}), WORD("bayes")}
    In case of list, it should extend list with extra items for accented and not accented versions
    """
    for group_label, pattern in rules:
        def gen():
            for p in pattern:
                (name, args, op) = p
                if name == "value":
                    (v1, v2) = (args, deaccent(args))
                    if v1 != v2:
                        yield ("any_of", (v1, v2,), op)
                    else:
                        yield p
                elif name == "any_of":
                    def items():
                        for w in args:
                            (v1, v2) = (w, deaccent(w))
                            if v1 != v2:
                                yield v1
                                yield v2
                            else:
                                yield v1

                    yield ("any_of", list(items()), op)
                else:
                    yield p

        yield (group_label, list(gen()))


def add_implicit_punct(rules, config):
    """
    When writing rule,
    user usually doesn't care about some punct characters between words.
    We add them implicitly (unless this setting is turned off)
    """
    for group_label, pattern in rules:
        def gen():
            for p in pattern:
                yield p
                yield ("punct", None, "?")

        if len(pattern) == 1:
            yield (group_label, pattern)
        else:
            yield (group_label, list(gen())[:-1])


def handle_multi_word(rules, config):
    """
    spaCy splits everything in tokens.
    Words with dash ends up in different tokens.

    We don't want for user to even care about this,
    so we make this work implicitly

    WORD("Knee-length") => WORD("Knee"), WORD("-"), WORD("length")
    """
    for group_label, pattern in rules:
        def gen():
            for p in pattern:
                (name, args, op) = p
                if name == "value" and is_complex(args):
                    yield ("phrase", args, op)
                else:
                    yield p

        yield (group_label, list(gen()))


def is_complex(arg):
    splitters = ["-", " "]
    return any([s in arg
                for s in splitters])


def has_complex(args):
    """
    Tells if any of arguments will be impacted by tokenizer
    """
    return any([is_complex(a)
                for a in args])


def branch_pattern(pattern, config):
    """
    Creates multiple lists for each possible permutation
    """
    root = Node()
    current = root
    depth = 0
    for idx, p in enumerate(pattern):
        if p[0] == "either":
            n = Node()
            current.add_next(n)
            current = n
            current.depth = depth
            for e in p[1]:
                current.add_child(e(config=config))
                depth += 1
        elif p[0] == "any_of" and has_complex(p[1]):
            _all = set(p[1])
            _complex = set(filter(is_complex, _all))
            simple = _all - _complex
            n = Node()
            current.add_next(n)
            current = n
            current.depth = depth
            current.add_child(("any_of", simple, p[2]))
            for c in sorted(_complex):
                current.add_child(("phrase", c, p[2]))
                depth += 1
        else:
            n = Node(p)
            current.add_next(n)
            current = n
            current.depth = depth

    for p in root.unwrap():
        yield p


def handle_rule_branching(rules, config):
    """
    If we have an OR statement, eg. `WORD(w1)|WORD(w2)`,
    Generic approach is to clone rules and use w1 in one, w2 in other.
    It may be an overkill, but some situations are not covered
    with simple approach
    """
    for group_label, pattern in rules:
        # Covering WORD(w1)|WORD(w2) case
        if any([p == "either"
                for (p, _, _) in pattern]):
            for p in branch_pattern(pattern, config):
                yield (group_label, p)

        # Covering case when there are complex items in list
        elif any([p == "any_of" and has_complex(o)
                  for (p, o, _) in pattern]):
            for p in branch_pattern(pattern, config):
                yield (group_label, p)
        else:
            yield (group_label, pattern)


def dummy(rules, config):
    """
    Placeholder which does nothing
    """
    logger.debug("Initial rules: {}".format(rules))
    return rules


def rule_tuple(d):
    return (d["label"], d["data"])


def preprocess_rules(root, config):
    logger.info("Preprocessing rules")

    rules = [rule_tuple(doc())
             for doc in root
             if doc and doc()]

    pipeline = [dummy, handle_deaccent, handle_rule_branching, handle_multi_word, handle_prefix]

    if config.implicit_punct:
        logger.info("Adding implicit Punctuations")
        pipeline.append(add_implicit_punct)

    return reduce(lambda acc, p: p(acc, config), pipeline, rules)



import logging
import types

from itertools import chain

logger = logging.getLogger(__name__)


def flatten(lst):
    if len(lst) > 1:
        return lst

    def explode(v):
        if callable(v):
            return v()
        else:
            return v

    new_lst = map(explode, lst)
    return chain(*new_lst)


def resolve_value(obj, config):
    context = []

    logger.debug("Resolving value: {0}".format(obj))

    if isinstance(obj, str):
        return obj

    elif isinstance(obj, tuple):
        return obj

    elif isinstance(obj, list):
        for item in obj:
            context.append(item)
        return context

    elif isinstance(obj, types.GeneratorType):
        return ("either", list(obj), None)

    return obj(config=config)


def ANY(config, op=None):
    return ("regex", r".*", op)


def PUNCT(config, op=None):
    return ("punct", None, op)


def MARK(type_, obj, config, op=None):
    return {
        "label": resolve_value(type_, config=config),
        "data": resolve_value(obj, config=config)
    }


def LOAD(*args, config):
    fpath = resolve_value(args[0], config=config)
    with open(fpath, "r") as f:
        return list([l.strip() for l in f.readlines()])


def ASSIGN(k, v, config, op=None):
    logger.debug("Assigning: {0} -> {1}".format(k, v))
    config.set_variable(k, resolve_value(v, config=config))


def IN_LIST(*args, config, op=None):
    variants = []
    for arg in flatten(args):
        variants.append(resolve_value(arg, config=config))
    return ("any_of", variants, None)


def PATTERN(*args, config, op=None):
    context = []
    for arg in args:
        result = resolve_value(arg, config=config)
        if isinstance(result, list):
            context += result
        else:
            context.append(result)

    return context


def WORD(*args, config, op=None):
    if len(args) == 1:
        literal = resolve_value(args[0], config=config)
        return ("value", literal, op)
    elif len(args) == 0:
        return ("regex", "(\\w|['_-])+", op)


def NUM(*args, config, op=None):
    if len(args) == 1:
        literal = resolve_value(args[0], config=config)
        return ("value", literal, op)
    elif len(args) == 0:
        return ("regex", "\\d+", op)


def POS(name, config, op=None):
    return ("pos", resolve_value(name, config=config), op)


def LEMMA(name, config, op=None):
    return ("lemma", resolve_value(name, config=config), op)


def ENTITY(name, config, op=None):
    return ("entity", resolve_value(name, config=config), op)


def PREFIX(name, config, op=None):
    return ("prefix", resolve_value(name, config=config), op)


def IMPORT(module, config):
    mod_name = resolve_value(module, config=config)
    config.register_module(mod_name)


def CONFIG(setting, value, config):
    logger.debug("Config {0} -> {1}".format(setting, value))
    config.set_config(setting, resolve_value(value, config=config))


def EXEC(obj, config):
    return resolve_value(obj, config=config)



import logging
import ply.lex as lex

logger = logging.getLogger(__name__)


class RitaLexer(object):
    tokens = [
        "KEYWORD",
        "LITERAL",
        "NAME",
        "LBRACKET",
        "RBRACKET",
        "LPAREN",
        "RPAREN",
        "ARROW",
        "COMMA",
        "MODIF_QMARK",
        "MODIF_STAR",
        "MODIF_PLUS",
        "ASSIGN",
        "EXEC",
        "PIPE",
    ]

    literals = ["{", "}", "(", ")", '"', ",", "=", "!", "|"]

    t_ignore = " \t"
    t_ignore_COMMENT = r"\#.*"
    t_ARROW = "->"
    t_LBRACKET = "{"
    t_RBRACKET = "}"
    t_LPAREN = r"\("
    t_RPAREN = r"\)"
    t_COMMA = ","
    t_MODIF_QMARK = r"\?"
    t_MODIF_STAR = r"\*"
    t_MODIF_PLUS = r"\+"
    t_EXEC = r"!"
    t_ASSIGN = r"="
    t_PIPE = r"\|"

    # Define a rule so we can track line numbers
    def t_newline(self, t):
        r"\n+"
        t.lexer.lineno += len(t.value)

    def t_KEYWORD(self, t):
        r"[A-Z_]{3,}"
        return t

    def t_LITERAL(self, t):
        r'".+?"'
        t.value = t.value.strip('"')
        return t

    def t_NAME(self, t):
        r"\w+"
        return t

    def t_error(self, t):
        logger.error("Invalid Token: {}".format(t.value[0]))
        t.lexer.skip(1)

    def build(self, **kwargs):
        self.lexer = lex.lex(module=self, errorlog=logger, **kwargs)
        return self.lexer

    def tokenize(self, data):
        self.lexer.input(data)
        while True:
            t = self.lexer.token()
            if t is None:
                break
            yield t



import logging

import ply.yacc as yacc

from functools import partial

from rita.lexer import RitaLexer
from rita import macros

logger = logging.getLogger(__name__)


def stub(*args, **kwargs):
    return None


def either(a, b):
    yield a
    yield b


def load_macro(name, config):
    try:
        return partial(getattr(macros, name), config=config)
    except Exception:
        pass

    def lazy_load(*args, **kwargs):
        print(config.modules)
        for mod in config.modules:
            try:
                fn = getattr(mod, name)
                return fn(*args, **kwargs)
            except Exception as ex:
                logger.error(ex)
                continue

        raise RuntimeError("MACRO {} not loaded".format(name))

    return lazy_load


def var_wrapper(variable, config):
    def wrapper(*args, **kwargs):
        logger.debug("Variables: {}".format(config.variables))
        return config.get_variable(variable)

    return wrapper


class RitaParser(object):
    tokens = RitaLexer.tokens
    precedence = (
        ("nonassoc", "ARROW"),
        ("nonassoc", "PIPE"),
        ("nonassoc", "COMMA"),
        ("left", "EXEC"),
        ("left", "ASSIGN"),
        ("left", "RBRACKET", "LBRACKET", "LPAREN", "RPAREN"),
        ("left", "KEYWORD", "NAME", "LITERAL"),
        ("right", "MODIF_QMARK", "MODIF_STAR", "MODIF_PLUS"),
    )

    def __init__(self, config):
        self.config = config

    def p_document(self, p):
        """
        DOCUMENT : MACRO_CHAIN
                 | MACRO_EXEC
                 | VARIABLE
        """
        logger.debug("Building initial document {}".format(p[1]))
        p[0] = [p[1]]

    def p_document_list(self, p):
        """
        DOCUMENT : DOCUMENT MACRO_CHAIN
                 | DOCUMENT MACRO_EXEC
                 | DOCUMENT VARIABLE
        """
        logger.debug("Extending document {}".format(p[2]))
        p[0] = p[1] + [p[2]]

    def p_macro_chain(self, p):
        " MACRO_CHAIN : MACRO ARROW MACRO "
        logger.debug("Have {0} -> {1}".format(p[1], p[3]))
        p[0] = partial(
            p[3],
            macros.PATTERN(p[1], config=self.config),
            config=self.config
        )

    def p_macro_chain_from_array(self, p):
        " MACRO_CHAIN : ARRAY ARROW MACRO "
        logger.debug("Have {0} -> {1}".format(p[1], p[3]))
        p[0] = partial(
            p[3],
            macros.PATTERN(*p[1], config=self.config),
            config=self.config
        )

    def p_macro_exec(self, p):
        " MACRO_EXEC : EXEC MACRO "
        logger.debug("Exec {0}".format(p[2]))
        macros.EXEC(p[2], config=self.config)
        p[0] = stub

    def p_macro_w_modif(self, p):
        """
        MACRO : MACRO MODIF_PLUS
              | MACRO MODIF_STAR
              | MACRO MODIF_QMARK
              | MACRO EXEC
        """
        logger.debug("Adding modifier to Macro {}".format(p[1]))
        fn = p[1]
        p[0] = partial(fn, op=p[2])

    def p_macro_wo_args(self, p):
        " MACRO : KEYWORD "
        fn = load_macro(p[1], config=self.config)
        logger.debug("Parsing macro (w/o args): {}".format(p[1]))
        p[0] = fn

    def p_macro_w_args(self, p):
        " MACRO : KEYWORD LPAREN ARGS RPAREN "
        logger.debug("Parsing macro: {0}, args: {1}".format(p[1], p[3]))
        fn = load_macro(p[1], config=self.config)
        p[0] = partial(fn, *p[3])

    def p_macro_from_array(self, p):
        " MACRO : KEYWORD ARRAY "
        logger.debug("Parsing macro: {0}, args: {1}".format(p[1], p[2]))
        fn = load_macro(p[1], config=self.config)
        p[0] = partial(fn, *p[2])

    def p_array(self, p):
        " ARRAY : LBRACKET ARGS RBRACKET "
        p[0] = p[2]

    def p_variable(self, p):
        " VARIABLE_NAME : NAME "
        p[0] = var_wrapper(p[1], self.config)

    def p_variable_from_args(self, p):
        " VARIABLE : NAME ASSIGN ARGS "
        if len(p[3]) == 1:
            macros.ASSIGN(p[1], p[3][0], config=self.config)
        else:
            macros.ASSIGN(p[1], p[3], config=self.config)

        p[0] = stub

    def p_either(self, p):
        " ARG : ARG PIPE ARG "
        p[0] = either(p[1], p[3])

    def p_arg_list(self, p):
        " ARGS : ARGS COMMA ARG "
        p[0] = p[1] + [p[3]]

    def p_args(self, p):
        " ARGS : ARG "
        p[0] = [p[1]]

    def p_arg(self, p):
        " ARG : LITERAL "
        p[0] = p[1]

    def p_arg_from_macro(self, p):
        " ARG : MACRO "
        p[0] = p[1]

    def p_arg_from_var(self, p):
        " ARG : VARIABLE_NAME "
        p[0] = p[1]()

    def p_arg_from_array(self, p):
        " ARGS : ARRAY "
        p[0] = p[1]

    def p_error(self, p):
        if p:
            logger.error("Syntax error at '{}'".format(p.value))
        else:
            logger.error("p is null")

    def build(self, **kwargs):
        self.lexer = RitaLexer().build(**kwargs)
        self.parser = yacc.yacc(module=self, errorlog=logger, **kwargs)

    def parse(self, data):
        if data.strip() == "":
            return []

        return self.parser.parse(data, lexer=self.lexer, debug=logger)



# coding: UTF-8
from setuptools import setup, find_packages

try:
    from pip._internal.req import parse_requirements
except:
    from pip.req import parse_requirements

from rita import get_version


install_reqs = list(parse_requirements("requirements/main.txt", session={}))
test_reqs = list(parse_requirements("requirements/test.txt", session={}))

with open("README.md", "r") as f:
    desc = f.read()

with open("CHANGELOG.md", "r") as f:
    desc += "## Changelog\n" + f.read()

setup(
    name="rita-dsl",
    version=get_version(),
    long_description=desc,
    long_description_content_type="text/markdown",
    description="DSL for building language rules",
    url="https://github.com/zaibacu/rita-dsl",
    author="Šarūnas Navickas",
    author_email="zaibacu@gmail.com",
    license="MIT",
    packages=find_packages(),
    install_requires=[str(ir.req) for ir in install_reqs],
    test_suite="pytest",
    tests_require=[str(tr.req) for tr in test_reqs],
    setup_requires=["pytest-runner"],
    entry_points={"console_scripts": ["rita = rita.run:main"]},
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
)



import rita


def setup_spacy(model, patterns=None, rules_path=None, rules_string=None, override_ents=True):
    from spacy.pipeline import EntityRuler
    ruler = EntityRuler(model, overwrite_ents=override_ents)
    if not patterns:
        if rules_path:
            patterns = rita.compile(rules_path, use_engine="spacy")
        elif rules_string:
            patterns = rita.compile_string(rules_string, use_engine="spacy")
        else:
            raise RuntimeError("Please provides rules. Either `patterns`, `rules_path` or `rules_string`")

        ruler.add_patterns(patterns)
    else:
        ruler.from_disk(patterns)

    model.add_pipe(ruler)
    return model
